#!/usr/bin/env python

from copy import copy, deepcopy
import random
import numpy as np
from scipy.stats import f

import lmfit
from lmfit import Parameter
from lmfit import (Parameters, Minimizer, conf_interval,
                   ci_report, conf_interval2d)

from lmfit.minimizer import MinimizerResult
from lmfit.model import (ModelResult, save_model, load_model,
                         save_modelresult, load_modelresult)
from lmfit.confidence import f_compare

from lmfit.printfuncs import gformat, getfloat_attr
from uncertainties import ufloat, correlated_values
from uncertainties import wrap as un_wrap

from ..symboltable import Group, isgroup


def isParameter(x):
    return (isinstance(x, Parameter) or
            x.__class__.__name__ == 'Parameter')

def param_value(val):
    "get param value -- useful for 3rd party code"
    while isinstance(val, Parameter):
        val = val.value
    return val

def format_param(par, length=10, with_initial=True):
    value = repr(par)
    if isParameter(par):
        value = gformat(par.value, length=length)
        if not par.vary and par.expr is None:
            value = f"{value} (fixed)"
        else:
            stderr = 'unknown'
            if par.stderr is not None:
                stderr = gformat(par.stderr, length=length)
                value = f"{value} +/-{stderr}"
                if par.vary and par.expr is None and with_initial:
                    value = f"{value} (init={gformat(par.init_value, length=length)})"
            if par.expr is not None:
                value = f"{value}  = '{par.expr}'"
    return value


def stats_table(results, labels=None, csv_output=False, csv_delim=','):
    """
    create a table comparing fit statistics for multiple fit results
    """
    stats = {'number of variables': 'nvarys',
             'chi-square': 'chi_square',
             'reduced chi-square': 'chi2_reduced',
             'r-factor': 'rfactor',
             'Akaike Info Crit': 'aic',
             'Bayesian Info Crit': 'bic'}

    nfits = len(results)
    if labels is not None:
        if len(labels) != len(results):
            raise ValueError('labels must be a list that is the same length as results')

    columns = [['Statistics']]
    if labels is None:
        labels = [f"  Fit {i+1}" for i in range(nfits)]
    for lab in labels:
        columns.append([lab])

    for sname, attr in stats.items():
        columns[0].append(sname)
        for i, result in enumerate(results):
            columns[i+1].append(getfloat_attr(result, attr))

    return format_table_columns(columns, csv_output=csv_output, csv_delim=csv_delim)

def paramgroups_table(pgroups, labels=None, csv_output=False, csv_delim=','):
    """
    create a table comparing parameters from a Feffit Parameter Grooup for multiple fit results
    """
    nfits = len(pgroups)
    if labels is not None:
        if len(labels) != len(pgroups):
            raise ValueError('labels must be a list that is the same length as Parameter Groups')

    columns = [['Parameter']]
    if labels is None:
        labels = [f" Fit {i+1}" for i in range(nfits)]
    for lab in labels:
        columns.append([lab])

    parnames = []
    for pgroup in pgroups:
        for pname in dir(pgroup):
            if pname not in parnames:
                parnames.append(pname)

    for pname in parnames:
        columns[0].append(pname)
        for i, pgroup in enumerate(pgroups):
            value = 'N/A'
            par = getattr(pgroup, pname, None)
            if par is not None:
                value = format_param(par, length=10, with_initial=False)
            columns[i+1].append(value)

    return format_table_columns(columns, csv_output=csv_output, csv_delim=csv_delim)


def format_table_columns(columns, csv_output=False, csv_delim=','):
    hjoin, rjoin, edge = '+', '|', '|'
    if csv_output:
       hjoin, rjoin, edge = csv_delim, csv_delim, ''

    ncols = len(columns)
    nrows = len(columns[0])
    slen = [2]*ncols
    for i, col in enumerate(columns):
        slen[i] = max(5, max([len(row) for row in col]))

    buff = []
    if not csv_output:
        header = edge + hjoin.join(['-'*(lx+2) for lx in slen]) + edge
        buff = [header]

    while len(columns[0]) > 0:
        values = [c.pop(0) for c in columns]
        row = rjoin.join([f" {l:{slen[i]}.{slen[i]}s} " for i, l in enumerate(values)])
        buff.append(edge + row + edge)
        if not csv_output and len(buff) == 2:
            buff.append(header)
    if not csv_output:
        buff.append(header)
    return '\n'.join(buff)


def f_test(ndata, nvars, chisquare, chisquare0, nfix=1):
    """return the F-test value for the following input values:
    f = f_test(ndata, nparams, chisquare, chisquare0, nfix=1)

    nfix = the number of fixed parameters.
    """
    return f_compare(ndata, nvars, chisquare, chisquare0, nfix=1)

def confidence_report(conf_vals, **kws):
    """return a formatted report of confidence intervals calcualted
    by confidence_intervals
    """
    return ci_report(conf_vals)

def asteval_with_uncertainties(*vals, **kwargs):
    """Calculate object value, given values for variables.

    This is used by the uncertainties package to calculate the
    uncertainty in an object even with a complicated expression.

    """
    _obj = kwargs.get('_obj', None)
    _pars = kwargs.get('_pars', None)
    _names = kwargs.get('_names', None)
    _asteval = _pars._asteval
    if (_obj is None or _pars is None or _names is None or
         _asteval is None or _obj._expr_ast is None):
        return 0
    for val, name in zip(vals, _names):
        _asteval.symtable[name] = val

    # re-evaluate all constraint parameters to
    # force the propagation of uncertainties
    [p._getval() for p in _pars.values()]
    return _asteval.eval(_obj._expr_ast)


wrap_ueval = un_wrap(asteval_with_uncertainties)


def eval_stderr(obj, uvars, _names, _pars):
    """Evaluate uncertainty and set ``.stderr`` for a parameter `obj`.

    Given the uncertain values `uvars` (list of `uncertainties.ufloats`),
    a list of parameter names that matches `uvars`, and a dictionary of
    parameter objects, keyed by name.

    This uses the uncertainties package wrapped function to evaluate the
    uncertainty for an arbitrary expression (in ``obj._expr_ast``) of
    parameters.

    """
    if not isinstance(obj, Parameter) or getattr(obj, '_expr_ast', None) is None:
        return
    uval = wrap_ueval(*uvars, _obj=obj, _names=_names, _pars=_pars)
    try:
        obj.stderr = uval.std_dev
    except Exception:
        obj.stderr = 0


class ParameterGroup(Group):
    """
    Group for Fitting Parameters
    """
    def __init__(self, name=None, **kws):
        if name is not None:
            self.__name__ = name
        if '_larch' in kws:
            kws.pop('_larch')
        self.__params__ = Parameters()
        Group.__init__(self)
        self.__exprsave__ = {}
        for key, val in kws.items():
            expr = getattr(val, 'expr', None)
            if expr is not None:
                self.__exprsave__[key] =  expr
                val.expr = None
            setattr(self, key, val)

        for key, val in self.__exprsave__.items():
            self.__params__[key].expr = val


    def __repr__(self):
        return '<Param Group {:s}>'.format(self.__name__)

    def __setattr__(self, name, val):
        if isParameter(val):
            if val.name != name:
                # allow 'a=Parameter(2, ..)' to mean Parameter(name='a', value=2, ...)
                nval = None
                try:
                    nval = float(val.name)
                except (ValueError, TypeError):
                    pass
                if nval is not None:
                    val.value = nval
            skip = getattr(val, 'skip', None)
            self.__params__.add(name, value=val.value, vary=val.vary, min=val.min,
                              max=val.max, expr=val.expr, brute_step=val.brute_step)
            val = self.__params__[name]

            val.skip = skip
        elif hasattr(self, '__params__') and not name.startswith('__'):
            self.__params__._asteval.symtable[name] = val
        self.__dict__[name] = val

    def __delattr__(self, name):
        self.__dict__.pop(name)
        if name in self.__params__:
            self.__params__.pop(name)

    def __add(self, name, value=None, vary=True, min=-np.inf, max=np.inf,
              expr=None, stderr=None, correl=None, brute_step=None, skip=None):
        if expr is None and isinstance(value, str):
            expr = value
            value = None
        if self.__params__  is not None:
            self.__params__.add(name, value=value, vary=vary, min=min, max=max,
                              expr=expr, brute_step=brute_step)
            self.__params__[name].stderr = stderr
            self.__params__[name].correl = correl
            self.__params__[name].skip = skip
            self.__dict__[name] = self.__params__[name]


def param_group(**kws):
    "create a parameter group"
    return ParameterGroup(**kws)

def randstr(n):
    return ''.join([chr(random.randint(97, 122)) for i in range(n)])

class unnamedParameter(Parameter):
    """A Parameter that can be nameless"""
    def __init__(self, name=None, value=None, vary=True, min=-np.inf, max=np.inf,
                 expr=None, brute_step=None, user_data=None, skip=None):
        if name is None:
            name = randstr(8)
        self.name = name
        self.user_data = user_data
        self.init_value = value
        self.min = min
        self.max = max
        self.brute_step = brute_step
        self.vary = vary
        self.skip = skip
        self._expr = expr
        self._expr_ast = None
        self._expr_eval = None
        self._expr_deps = []
        self._delay_asteval = False
        self.stderr = None
        self.correl = None
        self.from_internal = lambda val: val
        self._val = value
        self._init_bounds()
        Parameter.__init__(self, name, value=value, vary=vary,
                           min=min, max=max, expr=expr,
                           brute_step=brute_step,
                           user_data=user_data)

def param(*args, **kws):
    "create a fitting Parameter as a Variable"
    if len(args) > 0:
        a0 = args[0]
        if isinstance(a0, str):
            kws.update({'expr': a0})
        elif isinstance(a0, (int, float)):
            kws.update({'value': a0})
        else:
            raise ValueError("first argument to param() must be string or number")
        args = args[1:]
    if '_larch' in kws:
        kws.pop('_larch')
    if 'vary' not in kws:
        kws['vary'] = False

    return unnamedParameter(*args, **kws)

def guess(value,  **kws):
    """create a fitting Parameter as a Variable.
    A minimum or maximum value for the variable value can be given:
       x = guess(10, min=0)
       y = guess(1.2, min=1, max=2)
    """
    kws.update({'vary':True})
    return param(value, **kws)

def is_param(obj):
    """return whether an object is a Parameter"""
    return isParameter(obj)

def dict2params(pars):
    """sometimes we get a plain dict of Parameters,
    with vals that are Parameter objects"""
    if isinstance(pars, Parameters):
        return pars
    out = Parameters()
    for key, val in pars.items():
        if isinstance(val, Parameter):
            out[key] = val
    return out

def group2params(paramgroup):
    """take a Group of Parameter objects (and maybe other things)
    and put them into a lmfit.Parameters, ready for use in fitting
    """
    if isinstance(paramgroup, Parameters):
        return paramgroup
    if isinstance(paramgroup, dict):
        params = Parameters()
        for key, val in paramgroup.items():
            if isinstance(val, Parameter):
                params[key] = val
        return params


    if isinstance(paramgroup, ParameterGroup):
        return paramgroup.__params__

    params = Parameters()
    if paramgroup is not None:
        for name in dir(paramgroup):
            par = getattr(paramgroup, name)
            if getattr(par, 'skip', None) not in (False, None):
                continue
            if isParameter(par):
                params.add(name, value=par.value, vary=par.vary,
                           min=par.min, max=par.max,
                           brute_step=par.brute_step)
            else:
                params._asteval.symtable[name] = par

        # now set any expression (that is, after all symbols are defined)
        for name in dir(paramgroup):
            par = getattr(paramgroup, name)
            if isParameter(par) and par.expr is not None:
                params[name].expr = par.expr

    return params

def params2group(params, paramgroup):
    """fill Parameter objects in paramgroup with
    values from lmfit.Parameters
    """
    _params = getattr(paramgroup, '__params__', None)
    for name, param in params.items():
        this = getattr(paramgroup, name, None)
        if isParameter(this):
            if _params is not None:
                _params[name] = this
            for attr in ('value', 'vary', 'stderr', 'min', 'max', 'expr',
                         'name', 'correl', 'brute_step', 'user_data'):
                setattr(this, attr, getattr(param, attr, None))
            if this.stderr is not None:
                try:
                    this.uvalue = ufloat(this.value, this.stderr)
                except:
                    pass


def minimize(fcn, paramgroup, method='leastsq', args=None, kws=None,
             scale_covar=True, iter_cb=None, reduce_fcn=None, nan_polcy='omit',
             **fit_kws):
    """
    wrapper around lmfit minimizer for Larch
    """
    if isinstance(paramgroup, ParameterGroup):
        params = paramgroup.__params__
    elif isgroup(paramgroup):
        params = group2params(paramgroup)
    elif isinstance(Parameters):
        params = paramgroup
    else:
        raise ValueError('minimize takes ParamterGroup or Group as first argument')

    if args is None:
        args = ()
    if kws is None:
        kws = {}

    def _residual(params):
        params2group(params, paramgroup)
        return fcn(paramgroup, *args,  **kws)

    fitter = Minimizer(_residual, params, iter_cb=iter_cb,
                       reduce_fcn=reduce_fcn, nan_policy='omit', **fit_kws)

    result = fitter.minimize(method=method)
    params2group(result.params, paramgroup)

    out = Group(name='minimize results', fitter=fitter, fit_details=result,
                chi_square=result.chisqr, chi_reduced=result.redchi)

    for attr in ('aic', 'bic', 'covar', 'params', 'nvarys',
                 'nfree', 'ndata', 'var_names', 'nfev', 'success',
                 'errorbars', 'message', 'lmdif_message', 'residual'):
        setattr(out, attr, getattr(result, attr, None))
    return out

def fit_report(fit_result, modelpars=None, show_correl=True, min_correl=0.1,
               sort_pars=True, **kws):
    """generate a report of fitting results
    wrapper around lmfit.fit_report

    The report contains the best-fit values for the parameters and their
    uncertainties and correlations.

    Parameters
    ----------
    fit_result : result from fit
       Fit Group output from fit, or lmfit.MinimizerResult returned from a fit.
    modelpars : Parameters, optional
       Known Model Parameters.
    show_correl : bool, optional
       Whether to show list of sorted correlations (default is True).
    min_correl : float, optional
       Smallest correlation in absolute value to show (default is 0.1).
    sort_pars : bool or callable, optional
       Whether to show parameter names sorted in alphanumerical order. If
       False, then the parameters will be listed in the order they
       were added to the Parameters dictionary. If callable, then this (one
       argument) function is used to extract a comparison key from each
       list element.

    Returns
    -------
    string
       Multi-line text of fit report.


    """
    result = getattr(fit_result, 'fit_details', fit_result)
    if isinstance(result,  MinimizerResult):
        return lmfit.fit_report(result, modelpars=modelpars,
                                show_correl=show_correl,
                                min_correl=min_correl, sort_pars=sort_pars)
    elif isinstance(result,  ModelResult):
        return result.fit_report(modelpars=modelpars,
                                 show_correl=show_correl,
                                 min_correl=min_correl, sort_pars=sort_pars)
    else:
        result = getattr(fit_result, 'params', fit_result)
        if isinstance(result,  Parameters):
            return lmfit.fit_report(result, modelpars=modelpars,
                                    show_correl=show_correl,
                                    min_correl=min_correl, sort_pars=sort_pars)
        else:
            try:
                result = group2params(fit_result)
                return lmfit.fit_report(result, modelpars=modelpars,
                                        show_correl=show_correl,
                                        min_correl=min_correl, sort_pars=sort_pars)
            except (ValueError, AttributeError):
                pass
    return "Cannot make fit report with %s" % repr(fit_result)


def confidence_intervals(fit_result, sigmas=(1, 2, 3), **kws):
    """calculate the confidence intervals from a fit
    for supplied sigma values

    wrapper around lmfit.conf_interval
    """
    fitter = getattr(fit_result, 'fitter', None)
    result = getattr(fit_result, 'fit_details', None)
    return conf_interval(fitter, result, sigmas=sigmas, **kws)

def chi2_map(fit_result, xname, yname, nx=21, ny=21, sigma=3, **kws):
    """generate a confidence map for any two parameters for a fit

    Arguments
    ==========
       minout   output of minimize() fit (must be run first)
       xname    name of variable parameter for x-axis
       yname    name of variable parameter for y-axis
       nx       number of steps in x [21]
       ny       number of steps in y [21]
       sigma    scale for uncertainty range [3]

    Returns
    =======
        xpts, ypts, map

    Notes
    =====
     1.  sigma sets the extent of values to explore:
              param.value +/- sigma * param.stderr
    """
    #
    fitter = getattr(fit_result, 'fitter', None)
    result = getattr(fit_result, 'fit_details', None)
    if fitter is None or result is None:
        raise ValueError("chi2_map needs valid fit result as first argument")

    x = result.params[xname]
    y = result.params[yname]
    xrange = (x.value + sigma * x.stderr, x.value - sigma * x.stderr)
    yrange = (y.value + sigma * y.stderr, y.value - sigma * y.stderr)

    return conf_interval2d(fitter, result, xname, yname,
                           limits=(xrange, yrange),
                           nx=nx, ny=ny, nsigma=2*sigma, **kws)

_larch_name = '_math'
exports = {'param': param,
           'guess': guess,
           'param_group': param_group,
           'confidence_intervals': confidence_intervals,
           'confidence_report': confidence_report,
           'f_test': f_test, 'chi2_map': chi2_map,
           'is_param': isParameter,
           'isparam': isParameter,
           'minimize': minimize,
           'ufloat': ufloat,
           'fit_report': fit_report,
           'Parameters': Parameters,
           'Parameter': Parameter,
           'lm_minimize': minimize,
           'lm_save_model': save_model,
           'lm_load_model': load_model,
           'lm_save_modelresult': save_modelresult,
           'lm_load_modelresult': load_modelresult,
           }

for name in ('BreitWignerModel', 'ComplexConstantModel',
             'ConstantModel', 'DampedHarmonicOscillatorModel',
             'DampedOscillatorModel', 'DoniachModel',
             'ExponentialGaussianModel', 'ExponentialModel',
             'ExpressionModel', 'GaussianModel', 'Interpreter',
             'LinearModel', 'LognormalModel', 'LorentzianModel',
             'MoffatModel', 'ParabolicModel', 'Pearson7Model',
             'PolynomialModel', 'PowerLawModel',
             'PseudoVoigtModel', 'QuadraticModel',
             'RectangleModel', 'SkewedGaussianModel',
             'StepModel', 'StudentsTModel', 'VoigtModel'):
    val = getattr(lmfit.models, name, None)
    if val is not None:
        exports[name] = val

_larch_builtins = {'_math': exports}



# panel_refl.py : mirror reflectivity simulation
import wx
import wx.lib.plot as plot
import os.path
import math
import numpy
#import fluo
import readf1f2a
import SimpleParratt

#import input_file

class Plot(wx.Dialog):
    def __init__(self, parent, id, title):
        wx.Dialog.__init__(self, parent, id, 'SimRefl: mirror reflectivity simulation', size=(450, 420))
        self.plot_on = 0
        LogList=['Yes', 'No']
        xUnitList=['m rad', 'Deg', 'qz']
        xcol2=180
        btnSim = wx.Button(self,  1, 'Simulate!', (30,20))
        wx.StaticText(self, -1, 'y-axis in log?', (xcol2, 5))
        self.cb0 = wx.ComboBox(self, 0, 'No', (xcol2, 25), (95, 25), LogList, wx.CB_DROPDOWN)
        wx.StaticText(self, -1, 'x-axis unit?', (xcol2, 50))
        self.cb1 = wx.ComboBox(self, 0, 'm rad', (xcol2, 70), (95, 25), xUnitList, wx.CB_DROPDOWN)
        wx.StaticText(self, -1, "incident x-ray energy (eV)", (5, 50))
        self.Text1 = wx.TextCtrl(self, 2, '7000.0', (10, 70))
        wx.StaticText(self, -1, "film material", (5, 100))
        self.Text2 = wx.TextCtrl(self, 3, 'Pt', (10, 120))
        wx.StaticText(self, -1, "film thickness (Angstrom)", (5, 150))
        self.Text3 = wx.TextCtrl(self, 4, '400.0', (10, 170))
        wx.StaticText(self, -1, "film density (g/cm^3)", (5, 200))
        self.Text4 = wx.TextCtrl(self, 5, '21.45', (10, 220))
        wx.StaticText(self, -1, "rms roughness for all interfaces (Angstrom)", (5, 250))
        self.Text5 = wx.TextCtrl(self, 6, '1.0', (10, 270))
        wx.StaticText(self, -1, "substrate material", (xcol2-5, 100))
        self.Text6 = wx.TextCtrl(self, 7, 'Si', (xcol2, 120))
        wx.StaticText(self, -1, "substrate density (g/cm^3)", (xcol2-5, 150))
        self.Text7 = wx.TextCtrl(self, 8, '2.33', (xcol2, 170))
        wx.StaticText(self, -1, "Include 50 Angstrom Cr underlayer?", (xcol2-5, 200))
        CrList=['Yes', 'No']
        self.cb2 = wx.ComboBox(self, 10, 'Yes', (xcol2, 220), (95, 25), CrList, wx.CB_DROPDOWN)
        wx.StaticText(self, -1, "define arbitrary layer structure", (5, 310))
        self.Text8 = wx.TextCtrl(self, 12, 'GaAs/GaAs(250)/Fe(40)/Au(20)', (10, 330), (300, 25))
        btnSim = wx.Button(self,  11, 'Simulate arbitrary structure.', (10,360))

        wx.EVT_BUTTON(self, 1, self.OnSimulate)
        wx.EVT_TEXT(self, 3, self.OnText2)
        wx.EVT_TEXT(self, 7, self.OnText6)
        wx.EVT_CLOSE(self, self.OnQuit)
        wx.EVT_BUTTON(self, 11, self.OnSimulate1)

    def OnText2(self, event):
        text=self.Text2.GetValue()
        #out=fluo.nominal_density(text)	 # will be 1.0 if not in dictionary
        inputfile='NominalDensity.txt'
        f=open(inputfile)
        lines=f.readlines()
        out=1.0
        for line in lines:
            if not line.startswith('#'):
                words=line.split(',')
                if text==words[0]:
                    temp=words[1].split()
                    out=temp[0]
        f.close()
        self.Text4.SetValue(str(out))

    def OnText6(self, event):
        text=self.Text6.GetValue()
        #out=fluo.nominal_density(text)	 # will be 1.0 if not in dictionary
        inputfile='NominalDensity.txt'
        f=open(inputfile)
        lines=f.readlines()
        out=1.0
        for line in lines:
            if not line.startswith('#'):
                words=line.split(',')
                if text==words[0]:
                    temp=words[1].split()
                    out=temp[0]
        f.close()
        self.Text7.SetValue(str(out))


    def OnSimulate(self, event):
        reload(SimpleParratt)
    # --- Retrive values from panel --------------
        # self.Text1
        text=self.Text1.GetValue()
        eV0=float(text)
        eV0_min=500
        if eV0<=eV0_min:    eV0=eV0_min+10.
        # self.Text2
        film_mat=self.Text2.GetValue()
        # self.Text3
        text=self.Text3.GetValue()
        film_thick=float(text)
        # self.Text4
        text=self.Text4.GetValue()
        film_den=float(text)
        # self.Text5
        text=self.Text5.GetValue()
        film_roughness=float(text)
        # self.Text6
        subs_mat=self.Text6.GetValue()
        # self.Text7
        text=self.Text7.GetValue()
        subs_den=float(text)
        # self.cb2
        text=self.cb2.GetValue()
        use_Cr=0
        if text=='Yes': use_Cr=1
        # self.cb1
        text=self.cb1.GetValue()
        xUnit='m rad'
        xUnit=text
        # self.cb0
        text=self.cb0.GetValue()
        use_Log=False
        if text=='Yes': use_Log=True
        if use_Cr==1:
            NumLayers=2
        else:
            NumLayers=1
    # --- make lists as reflectivity calculation input ----------
        out=readf1f2a.get_delta(film_mat, film_den, eV0)  # used to be fluo.
        delta_film=out[0]
        la_film=out[2]*10000.	      # absorption length in cm, convert to microns
        thc_film=(2.*delta_film)**(0.5)*180./math.pi
        print('%s %3.4f Degrees (%3.4f m radians)' % ('critical angle for film:', thc_film, thc_film*math.pi/180.*1000.))
        th_step=0.001
        th_range=numpy.arange(0.000, thc_film*3.0, th_step)  # angular range
        depths=[0.0]  # calculate efield intensity at air/film interface
        layer_mat=[];	layer_thick=[];	 layer_den=[];	 layer_rough=[]; layer_tag=[]
        # air
        layer_mat.append('He')
        layer_thick.append(0.0)			# in Angstroms
        #layer_den.append(0.0013)		# in g/cm^2, this will make
        layer_den.append(1e-10)			 # force this to be zero for vacuum
        layer_rough.append(film_roughness)	# in Angstroms
        layer_tag.append('air')			# layer name
        # film
        layer_mat.append(film_mat)
        layer_thick.append(film_thick)		# in Angstroms
        layer_den.append(film_den)		# in g/cm^2
        layer_rough.append(film_roughness)	# in Angstroms
        layer_tag.append(film_mat+'_layer')
        # Cr underlayer if use_Cr=1
        if (NumLayers==2):
            layer_mat.append('Cr')
            layer_thick.append(50.0)
            layer_den.append(7.19)
            layer_rough.append(film_roughness)
            layer_tag.append('Cr_layer')
        # Substrate
        layer_mat.append(subs_mat)
        layer_thick.append(100000.)	    # not necessary
        layer_den.append(subs_den)
        layer_rough.append(film_roughness)  # not necessary
        layer_tag.append(subs_mat+'_substrate')
        # print(critical angle and absorption length for film and substrate)
        print('energy(eV) \t layer \t critical_angle(Deg.) \t absorption_length(micron) \n')
        print('%4.1f \t %s \t %4.4f \t %4.4f\n' % (eV0, film_mat, thc_film, la_film))
        out=readf1f2a.get_delta(subs_mat, subs_den, eV0)
        delta_subs=out[0]
        la_subs=out[2]*10000.	      # absorption length in cm, convert it to microns
        thc_subs=(2.*delta_subs)**(0.5)*180./math.pi
        print('%4.1f \t %s \t %4.4f \t %4.4f\n' % (eV0, subs_mat, thc_subs, la_subs))
    # --- call reflectivity in SimpleParratt.py	 ------------------
        SimpleParratt.reflectivity(eV0, th_range, layer_mat, layer_thick, layer_den,\
            layer_rough, layer_tag, depths)
    # --- Plot ----------------------------------------------------
        if self.plot_on==1:
            print(self.plot_on)
            self.plot.Destroy()
            #self.frm.Destroy()
        frm = wx.Frame(self, 1, 'SimpleParratt.txt', size=(600,450))
        self.data=[]; self.data0=[]
        th0=0;	refl_half=1e6
        inputfile='SimpleParratt.txt'
        f=open(inputfile)
        lines=f.readlines()
        self.xMin=1e6; self.xMax=-1e6; self.yMin=1e6; self.yMax=-1e6
        for line in lines:
            if line.startswith('#'):	continue
            words=line.split()
            if xUnit=='m rad':
                xx=float(words[0])*math.pi/180.0*1000.	# convert deg to m rad
            if xUnit=='Deg':
                xx=float(words[0])
            if xUnit=='qz':
                ang=float(words[0])
                qz=4*math.pi*(eV0/1000.0/12.39842)*math.sin(ang*math.pi/180.0)
                xx=qz
            if self.xMin>xx: self.xMin=xx
            if self.xMax<=xx: self.xMax=xx
            yy=float(words[1])
            if self.yMin>yy: self.yMin=yy
            if self.yMax<=yy: self.yMax=yy
            self.data.append((xx, yy))
            self.data0.append((xx, 0.5))
            temp=abs(yy-0.5)
            if temp<=refl_half:
                refl_half=temp
                th0=xx
        self.yMax=1.0
        f.close()
        sample=''
        for (ii,item) in enumerate(layer_mat):
            if ii==0:
                sample='vacuum'
            else:
                sample=sample+'/'+item
        #title=str(eV0)+'eV, '+sample+', Refl.=0.5 at '+str(th0)+' m rad.'
        text0=' m rad.'
        text1='Incident angle (m rad)'
        if xUnit=='Deg':
            text0=' Deg'
            text1='Incident angle (Deg)'
        if xUnit=='qz':
            text0=' qz (1/Angstrom)'
            text1='qz (1/Agnstrom)'
        title='%4.1f, %s, %s %s %3.3f %s' % (eV0, 'eV', sample, 'Refl.=0.5 at', th0, text0)
        #------------------------------------------------
        client = plot.PlotCanvas(frm)
        # 9/1/2010: enable zooming function, double click for default size
        client.SetEnableZoom(True)
        line = plot.PolyLine(self.data, legend='', colour='red', width=1)
        line0 = plot.PolyLine(self.data0, legend='', colour='black', width=0.5)
        gc = plot.PlotGraphics([line, line0], title, text1, 'Reflectivity')
        client.setLogScale((False,False))
        if use_Log==True: client.setLogScale((False,True))
        client.Draw(gc,	 xAxis= (self.xMin,self.xMax), yAxis= (self.yMin,self.yMax))
        frm.Show(True)
        self.plot_on = 1
        self.plot=frm
        #------------------------------------------------

    def OnSimulate1(self, event):  # for arbitrary structure
        reload(SimpleParratt)
    # --- Retrive values from panel --------------
        # self.Text1
        text=self.Text1.GetValue()
        eV0=float(text)
        eV0_min=500
        if eV0<=eV0_min:    eV0=eV0_min+10.
        # self.Text2
        film_mat=self.Text2.GetValue()
        # self.Text3
        text=self.Text3.GetValue()
        film_thick=float(text)
        # self.Text4
        text=self.Text4.GetValue()
        film_den=float(text)
        # self.Text5
        text=self.Text5.GetValue()
        film_roughness=float(text)
        # self.Text6
        subs_mat=self.Text6.GetValue()
        # self.Text7
        text=self.Text7.GetValue()
        subs_den=float(text)
        # self.cb2
        text=self.cb2.GetValue()
        use_Cr=0
        if text=='Yes': use_Cr=1
        # self.cb1
        text=self.cb1.GetValue()
        xUnit='m rad'
        xUnit=text
        # self.cb0
        text=self.cb0.GetValue()
        use_Log=False
        if text=='Yes': use_Log=True
        if use_Cr==1:
            NumLayers=2
        else:
            NumLayers=1
        LayerStructure0=self.Text8.GetValue()
        import read_FilmName
        film1=read_FilmName.Film(LayerStructure0)
        film1.get_structure()
        film1.reverse_structure()  # reverse the layer order so that the top is vaccum
        layer_mat=[];	layer_thick=[];	 layer_den=[];	 layer_rough=[]; layer_tag=[]
        layer_thc=[]; layer_la=[]
        print('energy(eV) \t layer \t th_c(Deg.) \t th_c(m rad) \t absorption_length(micron) \t density (g/cc)\n')
        for (ii, layer1) in enumerate(film1.LayerList):
            layer_mat.append(str(layer1.composition))
            inputfile='NominalDensity.txt'
            f=open(inputfile)
            lines=f.readlines()
            out=1.0
            for line in lines:
                if not line.startswith('#'):
                    words=line.split(',')
                    if layer1.composition==words[0]:
                        temp=words[1].split()
                        out=temp[0]
            layer1.density=float(out)
            f.close()
            layer_den.append(float(layer1.density))
            layer_thick.append(float(layer1.thickness))
            layer_rough.append(float(layer1.rms))
            layer_tag.append(str(layer1.tag))
            out=readf1f2a.get_delta(layer1.composition, layer1.density, eV0)
            la_layer=out[2]*10000.0  # absorption length in cm, convert to microns
            delta_layer=out[0]
            thc_layer=(2.*delta_layer)**(0.5)*180./math.pi  # critical angle in deg
            if ii==1:
                thc_toplayer=thc_layer
            print('%4.1f \t %s \t %4.4f \t %4.4f \t %4.1f \t %s\n' % \)
                (eV0, layer1.composition+'-'+layer1.tag, thc_layer, thc_layer*math.pi/180.*1000., la_layer, layer1.density)
        th_step=0.001
        th_range=numpy.arange(0.000, thc_toplayer*12.0, th_step)  # angular range
        depths=[0.0]  # calculate efield intensity at air/film interface
        #print(layer_mat)
        #print(layer_thick)
        #print(layer_den)
        #print(layer_rough)
        #layer_mat=['He', 'Cr', 'Si']; layer_thick=[0, 10., 10000]; layer_den=[1,1,1]; layer_rough=[1,1,1]
    # --- call reflectivity in SimpleParratt.py	 ------------------
        SimpleParratt.reflectivity(eV0, th_range, layer_mat, layer_thick, layer_den,\
            layer_rough, layer_tag, depths)
    # --- Plot ----------------------------------------------------
        if self.plot_on==1:
            print(self.plot_on)
            self.plot.Destroy()
            #self.frm.Destroy()
        frm = wx.Frame(self, 1, 'SimpleParratt.txt', size=(600,450))
        self.data=[]; self.data0=[]
        th0=0;	refl_half=1e6
        inputfile='SimpleParratt.txt'
        f=open(inputfile)
        lines=f.readlines()
        self.xMin=1e6; self.xMax=-1e6; self.yMin=1e6; self.yMax=-1e6
        for line in lines:
            if line.startswith('#'):	continue
            words=line.split()
            if xUnit=='m rad':
                xx=float(words[0])*math.pi/180.0*1000.	# convert deg to m rad
            if xUnit=='Deg':
                xx=float(words[0])
            if xUnit=='qz':
                ang=float(words[0])
                qz=4*math.pi*(eV0/1000.0/12.39842)*math.sin(ang*math.pi/180.0)
                xx=qz
            if self.xMin>xx: self.xMin=xx
            if self.xMax<=xx: self.xMax=xx
            yy=float(words[1])
            if self.yMin>yy: self.yMin=yy
            if self.yMax<=yy: self.yMax=yy
            self.data.append((xx, yy))
            self.data0.append((xx, 0.5))
            temp=abs(yy-0.5)
            if temp<=refl_half:
                refl_half=temp
                th0=xx
        self.yMax=1.0
        f.close()
        sample=''
        for (ii,item) in enumerate(layer_mat):
            if ii==0:
                sample='vacuum'
            else:
                sample=sample+'/'+item
        #title=str(eV0)+'eV, '+sample+', Refl.=0.5 at '+str(th0)+' m rad.'
        text0=' m rad.'
        text1='Incident angle (m rad)'
        if xUnit=='Deg':
            text0=' Deg'
            text1='Incident angle (Deg)'
        if xUnit=='qz':
            text0=' qz (1/Angstrom)'
            text1='qz (1/Agnstrom)'
        title='%4.1f, %s, %s %s %3.3f %s' % (eV0, 'eV', sample, 'Refl.=0.5 at', th0, text0)
        #------------------------------------------------
        client = plot.PlotCanvas(frm)
        # 9/1/2010: enable zooming function, double click for default size
        client.SetEnableZoom(True)
        line = plot.PolyLine(self.data, legend='', colour='red', width=1)
        line0 = plot.PolyLine(self.data0, legend='', colour='black', width=0.5)
        gc = plot.PlotGraphics([line, line0], title, text1, 'Reflectivity')
        client.setLogScale((False,False))
        if use_Log==True: client.setLogScale((False,True))
        client.Draw(gc,	 xAxis= (self.xMin,self.xMax), yAxis= (self.yMin,self.yMax))
        frm.Show(True)
        self.plot_on = 1
        self.plot=frm
        #------------------------------------------------


    def OnQuit(self, event):
        self.plot_on=0
        self.Destroy()





from time import time
import numpy as np
fft_numpy =  np.fft.fft

fft_mkln = fft_mkls = fft_ws = fft_wp = fft_wn = None

try:
    import mkl_fft
    from mkl_fft.interfaces.numpy_fft import fft as fft_mkln
    from mkl_fft.interfaces.scipy_fft import fft as fft_mkls
except ImportError:
    print("mkl not available")

try:
    from pyfftw.interfaces.scipy_fft import fft as fft_ws
    from pyfftw.interfaces.scipy_fftpack import fft as fft_wp
    from pyfftw.interfaces.numpy_fft import fft as fft_wn
except:
    print("pyfftw not available")

from scipy.fftpack import fft as fft_pack
from scipy.fft import fft as fft_scipy

from larch.io import read_ascii
from larch.xafs import pre_edge, autobk, xftf, xftf_prep


METHODS = {}
for key, val in {'numpy': fft_numpy, 'scipy': fft_scipy,
                 'fftpack': fft_pack,
                 'mkl_numpy': fft_mkln, 'mkl_scipy': fft_mkls,
                 'fftw_numpy': fft_wn, 'fftw_scipy': fft_ws,
                 'fftw_pack': fft_wp}.items():

    if callable(val):
        METHODS[key] = val

fname = '../xafsdata/feo_rt1.xdi'
dat = read_ascii(fname, labels='energy mu i0')

pre_edge(dat)
autobk(dat, rbkg=0.9, kweight=2)

xftf(dat, kmin=2, kmax=13, dk=3, window='hanning', kweight=2)

# plot_chir(dat, win=2)


xk = dat.k

xchi, win = xftf_prep(dat.k, dat.chi, kmin=2, kmax=13, dk=3, dk2=3,
                      kstep=0.05, window='hanning', kweight=2)

nfft = 2048
kstep = 0.05
rstep = np.pi /(nfft*kstep)

cchi = np.zeros(nfft, dtype='complex128')
cchi[0:len(xchi)] = xchi*win
r    = rstep * np.arange(nfft//2)

rscale = kstep/np.sqrt(np.pi)
results = {}
for key in METHODS:
    results[key] = None

out = {}
for key in METHODS:
    out[key] = []

#warm up
for name, meth in METHODS.items():
    for i in range(5):
        t0 = time()
        chir = rscale * meth(cchi)
        results[name] = chir.real

for name, meth in METHODS.items():
    for i in range(50):
        t0 = time()
        for j in range(25):
            chir = rscale * meth(cchi)
        out[name].append(1e4*(time() -t0))


k,m,s,x = 'Method', 'Mean', 'Std', 'Max'
print(f"{k:10s} :  {m:8s}   {s:8s} {x:8s}")
for k, v in out.items():
    vn = np.array(v)
    print(f"{k:10s} :  {vn.mean():8.3f}   {vn.std():8.3f} {vn.max():8.3f}")

print("done")



import os
from pathlib import Path
from datetime import datetime
from collections import namedtuple
from larch.utils import read_textfile

FPathInfo = namedtuple('FeffPathInfo',
                       ('filename', 'absorber', 'shell', 'reff', 'nleg',
                        'degen', 'cwratio', 'geom', 'geometry'))
class FeffCalcResults:
    def __init__(self, folder=None, header=None, ipots=None,
                 paths=None, datetime=None, absorber=None,
                 shell=None, input_text=None):
        self.folder = Path(folder).absolute().as_posix()
        self.header = header
        self.ipots = ipots
        self.paths = paths
        self.datetime = datetime
        self.absorber = absorber
        self.shell = shell
        self.input_text = input_text

    def __getstate__(self):
        """Get state for pickle."""
        return (self.folder, self.absorber, self.shell, self.header,
                self.ipots, self.paths, self.datetime, self.input_text)

    def __setstate__(self, state):
        """Set state from pickle."""
        (self.folder, self.absorber, self.shell, self.header,
         self.ipots, self.paths, self.datetime, self.input_text) = state


def get_feff_pathinfo(folder):
    """get list of Feff path info for a Feff folder
    """
    fdat = Path(folder, 'files.dat')
    pdat = Path(folder, 'paths.dat')
    f001 = Path(folder, 'feff0001.dat')
    finp = Path(folder, 'feff.inp')

    # check for valid, complete calculation
    if (not fdat.exists() or not pdat.exists() or
        not f001.exists() or not finp.exists()):
        return FeffCalcResults(folder,  absorber=None,
                               shell=None, ipots=[], header='',
                               paths=[], datetime=None)


    dtime = datetime.fromtimestamp(os.stat(fdat).st_mtime).isoformat()
    fdatlines = read_textfile(fdat).split('\n')
    pathslines = read_textfile(pdat).split('\n')

    xpaths = {}
    paths = {}
    geoms = {}
    header = []
    shell = 'K'
    waiting_for_dashes = True
    for xline in fdatlines:
        xline = xline.strip()
        if xline.startswith('----'):
            waiting_for_dashes = False
        if waiting_for_dashes:
            header.append(xline)

            if (xline.startswith('Abs ') and 'Rmt=' in xline and
                'Rnm=' in xline and 'shell' in xline):
                words = xline.replace('shell', '').strip().split()
                shell = words[-1]
            continue
        if xline.startswith('feff0'):
            w = xline.split()
            index = int(w[0].replace('feff', '').replace('.dat', ''))
            paths[index] = [w[0], w[5], w[4], w[3], w[2]]
            geoms[index] = []

    ipots = ['']*12
    waiting_for_dashes = True
    for i, xline in enumerate(pathslines):
        xline = xline.strip()
        if xline.startswith('----'):
            waiting_for_dashes = False
        if waiting_for_dashes:
            continue
        if 'index, nleg' in xline:
            words = xline.split()
            index = int(words[0])
            nleg  = int(words[1])
            elems = []
            pgeom = []
            for j in range(nleg+1):
                xline = pathslines[j+i+1].strip().replace("'", '')
                if xline.startswith('x   '):
                    continue
                words = xline.split()
                atname = words[4].strip()
                ipot = int(words[3])
                if ipot not in ipots:
                    ipots[ipot] = atname
                elems.append(ipot)
                r, x, y, z, beta, eta = [float(words[i]) for i in (5, 0, 1, 2, 6, 7)]
                pgeom.append((atname, ipot, r, x, y, z, beta, eta))
                if j == nleg:
                    pgeom.insert(0, (atname, ipot, r, x, y, z, beta, eta))
            if index in paths:
                paths[index].append(elems)
                geoms[index] = pgeom

    ipots = [i for i in ipots if len(i) > 0]
    absorber = ipots[0]
    ipots[0] = '[%s]' % ipots[0]
    opaths = []
    for pindex, pinfo in paths.items():
        pots = [0] + pinfo[5]
        geom =  ' > '.join([ipots[i] for i in pots])
        opaths.append(FPathInfo(filename=pinfo[0],
                                absorber=absorber,
                                shell=shell,
                                reff=float(pinfo[1]),
                                nleg=int(float(pinfo[2])),
                                degen=float(pinfo[3]),
                                cwratio=float(pinfo[4]),
                                geom=geom,
                                geometry=geoms[pindex]))


    # read absorbing shell
    for line in header:
        line = line.strip()

    # read input
    try:
        input_text = read_textfile(fin)
    except:
        input_text = '<not available>'

    return FeffCalcResults(folder,
                           absorber=absorber,
                           shell=shell,
                           ipots=ipots,
                           header='\n'.join(header),
                           paths=opaths,
                           input_text=input_text,
                           datetime=dtime)



#!/usr/bin/env python
"""
 uncategorized XY data Panel
"""
import time
import wx
import numpy as np

from functools import partial
from xraydb import guess_edge, atomic_number

from larch.utils import gformat, path_split
from larch.math import index_of
from larch.xafs.xafsutils import guess_energy_units

from larch.wxlib import (BitmapButton, FloatCtrl, FloatSpin, get_icon,
                         SimpleText, pack, Button, HLine, Choice, Check,
                         GridPanel, CEN, RIGHT, LEFT, plotlabels,
                         get_zoomlimits, set_zoomlimits)

from larch.utils.strutils import fix_varname, fix_filename, file2groupname

from larch.utils.physical_constants import ATOM_NAMES
from larch.wxlib.plotter import last_cursor_pos
from .taskpanel import TaskPanel, autoset_fs_increment, update_confval
from .config import (make_array_choice, EDGES, ATSYMS,
                     NNORM_CHOICES, NNORM_STRINGS, NORM_METHODS)

np.seterr(all='ignore')

PLOTOPTS_1 = dict(style='solid', marker='None')
PLOTOPTS_2 = dict(style='short dashed', zorder=3, marker='None')
PLOTOPTS_D = dict(style='solid', zorder=2, side='right', marker='None')

# Plot_EnergyRanges = {'full X range': None }

PlotOne_Choices = {'XY Data': 'y',
                   'Scaled Data': 'ynorm',
                   'Derivative ': 'dydx',
                   'XY Data + Derivative': 'y+dydx',
                   'Scaled Data + Derivative': 'ynorm+dydx',
                   }

PlotSel_Choices = {'XY Data': 'y',
                   'Scaled Data': 'ynorm',
                   'Derivative': 'dydx'}

FSIZE = 120
FSIZEBIG = 175


class XYDataPanel(TaskPanel):
    """XY Data Panel"""
    def __init__(self, parent, controller=None, **kws):
        TaskPanel.__init__(self, parent, controller, panel='xydata', **kws)

    def build_display(self):
        panel = self.panel
        self.wids = {}
        self.last_plot_type = 'one'

        trow = wx.Panel(panel)
        plot_sel = Button(trow, 'Plot Selected Groups', size=(175, -1),
                          action=self.onPlotSel)
        plot_one = Button(trow, 'Plot Current Group', size=(175, -1),
                          action=self.onPlotOne)

        self.plotsel_op = Choice(trow, choices=list(PlotSel_Choices.keys()),
                                 action=self.onPlotSel, size=(300, -1))
        self.plotone_op = Choice(trow, choices=list(PlotOne_Choices.keys()),
                                 action=self.onPlotOne, size=(300, -1))

        opts = {'digits': 2, 'increment': 0.05, 'value': 0, 'size': (FSIZE, -1)}
        plot_voff = self.add_floatspin('plot_voff', with_pin=False,
                                       parent=trow,
                                       action=self.onVoffset,
                                       max_val=10000, min_val=-10000,
                                       **opts)

        vysize, vxsize = plot_sel.GetBestSize()
        voff_lab = wx.StaticText(parent=trow, label='  Y Offset:', size=(80, vxsize),
                                 style=wx.RIGHT|wx.ALIGN_CENTRE_HORIZONTAL|wx.ST_NO_AUTORESIZE)

        self.plotone_op.SetSelection(0)
        self.plotsel_op.SetSelection(1)

        tsizer = wx.GridBagSizer(3, 3)
        tsizer.Add(plot_sel,        (0, 0), (1, 1), LEFT, 2)
        tsizer.Add(self.plotsel_op, (0, 1), (1, 1), LEFT, 2)
        tsizer.Add(voff_lab,        (0, 2), (1, 1), RIGHT, 2)
        tsizer.Add(plot_voff,       (0, 3), (1, 1), RIGHT, 2)
        tsizer.Add(plot_one,       (1, 0), (1, 1), LEFT, 2)
        tsizer.Add(self.plotone_op, (1, 1), (1, 1), LEFT, 2)

        pack(trow, tsizer)

        scale = self.add_floatspin('scale', action=self.onSet_Scale,
                                                digits=6, increment=0.05, value=1.0,
                                                size=(FSIZEBIG, -1))

        xshift = self.add_floatspin('xshift', action=self.onSet_XShift,
                                                digits=6, increment=0.05, value=0.0,
                                                size=(FSIZEBIG, -1))

        self.wids['is_frozen'] = Check(panel, default=False, label='Freeze Group',
                                       action=self.onFreezeGroup)

        use_auto = Button(panel, 'Use Default Settings', size=(200, -1),
                          action=self.onUseDefaults)

        def CopyBtn(name):
            return Button(panel, 'Copy', size=(60, -1),
                          action=partial(self.onCopyParam, name))

        copy_all = Button(panel, 'Copy All Parameters', size=(200, -1),
                          action=partial(self.onCopyParam, 'all'))

        add_text = self.add_text
        HLINEWID = 700

        panel.Add(SimpleText(panel, 'XY Data, General',
                             size=(650, -1), **self.titleopts), style=LEFT, dcol=4)
        panel.Add(trow, dcol=4, newrow=True)

        panel.Add(HLine(panel, size=(HLINEWID, 3)), dcol=4, newrow=True)

        add_text('XY Data:')
        panel.Add(use_auto, dcol=1)
        panel.Add(SimpleText(panel, 'Copy to Selected Groups:'), style=RIGHT, dcol=2)

        add_text('Scale Factor:')
        panel.Add(scale)
        panel.Add(SimpleText(panel, 'Scaled Data = Y /(Scale Factor)'))
        panel.Add(CopyBtn('scale'), dcol=1, style=RIGHT)

        add_text('Shift X scale:' )
        panel.Add(xshift)
        panel.Add(CopyBtn('xshift'), dcol=2, style=RIGHT)

        panel.Add(HLine(panel, size=(HLINEWID, 3)), dcol=4, newrow=True)
        panel.Add(self.wids['is_frozen'], newrow=True)
        panel.Add(copy_all, dcol=3, style=RIGHT)

        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((5, 5), 0, LEFT, 3)
        sizer.Add(panel, 0, LEFT, 3)
        sizer.Add((5, 5), 0, LEFT, 3)
        pack(self, sizer)

    def get_config(self, dgroup=None):
        """custom get_config"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        if dgroup is None:
            return self.get_defaultconfig()
        self.read_form()

        defconf = self.get_defaultconfig()
        conf = getattr(dgroup.config, self.configname, defconf)

        for k, v in defconf.items():
            if k not in conf:
                conf[k] = v

        fname = getattr(dgroup, 'filename', None)
        if fname is None:
            fname = getattr(dgroup, 'groupname', None)
            if fname is None:
                fname =file2groupname('unknown_group',
                                      symtable=self._larch.symtable)

        for attr in ('scale', 'xshift'):
            conf[attr] = getattr(dgroup, attr, conf[attr])

        setattr(dgroup.config, self.configname, conf)
        return conf

    def fill_form(self, dgroup):
        """fill in form from a data group"""
        opts = self.get_config(dgroup)
        self.skip_process = True

        self.plotone_op.SetChoices(list(PlotOne_Choices.keys()))
        self.plotsel_op.SetChoices(list(PlotSel_Choices.keys()))

        self.wids['scale'].SetValue(opts['scale'])
        self.wids['xshift'].SetValue(opts['xshift'])

        frozen = opts.get('is_frozen', False)
        frozen = getattr(dgroup, 'is_frozen', frozen)

        self.wids['is_frozen'].SetValue(frozen)
        self._set_frozen(frozen)
        wx.CallAfter(self.unset_skip_process)

    def unset_skip_process(self):
        self.skip_process = False

    def read_form(self):
        "read form, return dict of values"
        form_opts = {}
        form_opts['scale'] = self.wids['scale'].GetValue()
        form_opts['xshift'] = self.wids['xshift'].GetValue()
        return form_opts


    def _set_frozen(self, frozen):
        try:
            dgroup = self.controller.get_group()
            dgroup.is_frozen = frozen
        except:
            pass

        for wattr in ('scale',):
            self.wids[wattr].Enable(not frozen)

    def onFreezeGroup(self, evt=None):
        self._set_frozen(evt.IsChecked())

    def onPlotEither(self, evt=None):
        if self.last_plot_type == 'multi':
            self.onPlotSel(evt=evt)
        else:
            self.onPlotOne(evt=evt)

    def onPlotOne(self, evt=None):
        self.last_plot_type = 'one'
        self.plot(self.controller.get_group())
        wx.CallAfter(self.controller.set_focus)

    def onVoffset(self, evt=None):
        time.sleep(0.002)
        wx.CallAfter(self.onPlotSel)

    def onPlotSel(self, evt=None):
        newplot = True
        self.last_plot_type = 'multi'
        group_ids = self.controller.filelist.GetCheckedStrings()
        if len(group_ids) < 1:
            return
        last_id = group_ids[-1]

        groupname = self.controller.file_groups[str(last_id)]
        dgroup = self.controller.get_group(groupname)

        plot_choices = PlotSel_Choices

        ytitle = self.plotsel_op.GetStringSelection()
        yarray_name = plot_choices.get(ytitle, 'ynorm')
        ylabel = getattr(plotlabels, yarray_name, ytitle)
        xlabel = getattr(dgroup, 'plot_xlabel', getattr(plotlabels, 'xplot'))

        voff = self.wids['plot_voff'].GetValue()
        plot_traces = []
        newplot = True
        plotopts = self.controller.get_plot_conf()
        popts = {'style': 'solid', 'marker': None}
        popts['linewidth'] = plotopts.pop('linewidth')
        popts['marksize'] = plotopts.pop('markersize')
        popts['grid'] = plotopts.pop('show_grid')
        popts['fullbox'] = plotopts.pop('show_fullbox')

        for ix, checked in enumerate(group_ids):
            groupname = self.controller.file_groups[str(checked)]
            dgroup = self.controller.get_group(groupname)
            if dgroup is None:
                continue
            trace = {'xdata': dgroup.xplot,
                     'ydata': getattr(dgroup, yarray_name) + ix*voff,
                     'label': dgroup.filename, 'new': newplot}
            trace.update(popts)
            plot_traces.append(trace)
            newplot = False

        ppanel = self.controller.get_display(stacked=False).panel
        zoom_limits = get_zoomlimits(ppanel, dgroup)

        nplot_traces = len(ppanel.conf.traces)
        nplot_request = len(plot_traces)
        if nplot_request > nplot_traces:
            linecolors = ppanel.conf.linecolors
            ncols = len(linecolors)
            for i in range(nplot_traces, nplot_request+5):
                ppanel.conf.init_trace(i,  linecolors[i%ncols], 'dashed')


        ppanel.plot_many(plot_traces, xlabel=plotlabels.xplot, ylabel=ylabel,
                         zoom_limits=zoom_limits, show_legend=True)
        set_zoomlimits(ppanel, zoom_limits) or ppanel.unzoom_all()
        ppanel.canvas.draw()
        wx.CallAfter(self.controller.set_focus)

    def onUseDefaults(self, evt=None):
        self.wids['scale'].SetValue(1.0)
        self.wids['xshift'].SetValue(0.0)


    def onCopyAuto(self, evt=None):
        opts = dict(scale=1)
        for checked in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(checked)]
            grp = self.controller.get_group(groupname)
            if grp != self.controller.group and not getattr(grp, 'is_frozen', False):
                self.update_config(opts, dgroup=grp)
                self.fill_form(grp)
                self.process(grp, force=True)


    def onSaveConfigBtn(self, evt=None):
        conf = self.get_config()
        conf.update(self.read_form())


    def onCopyParam(self, name=None, evt=None):
        conf = self.get_config()
        form = self.read_form()
        conf.update(form)
        dgroup = self.controller.get_group()
        self.update_config(conf)
        self.fill_form(dgroup)
        opts = {}
        name = str(name)
        def copy_attrs(*args):
            for a in args:
                opts[a] = conf[a]
        if name == 'all':
            copy_attrs('scale')
        elif name == 'scale':
            copy_attrs('scale')

        for checked in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(checked)]
            grp = self.controller.get_group(groupname)
            if grp != self.controller.group and not getattr(grp, 'is_frozen', False):
                self.update_config(opts, dgroup=grp)
                for key, val in opts.items():
                    if hasattr(grp, key):
                        setattr(grp, key, val)
                self.fill_form(grp)
                self.process(grp, force=True)

    def onSet_Scale(self, evt=None, value=None):
        "handle setting scale"
        scale = self.wids['scale'].GetValue()
        if scale < 0:
            self.wids['scale'].SetValue(abs(scale))
        self.update_config({'scale': self.wids['scale'].GetValue()})
        autoset_fs_increment(self.wids['scale'], abs(scale))
        time.sleep(0.01)
        wx.CallAfter(self.onReprocess)

    def onSet_XShift(self, evt=None, value=None):
        "handle x shift"
        xshift = self.wids['xshift'].GetValue()
        self.update_config({'xshift': self.wids['xshift'].GetValue()})
        autoset_fs_increment(self.wids['xshift'], abs(xshift))
        time.sleep(0.01)
        wx.CallAfter(self.onReprocess)

    def pin_callback(self, opt='__', xsel=None, relative_e0=True, **kws):
        """
        get last selected point from a specified plot window
        and fill in the value for the widget defined by `opt`.

        by default it finds the latest cursor position from the
        cursor history of the first 20 plot windows.
        """
        if xsel is None or opt not in self.wids:
            return
        if opt == 'scale':
            self.wids['scale'].SetValue(kws['ysel'])
        elif opt == 'xshift':
            self.wids['xshift'].SetValue(kws['xsel'])
        time.sleep(0.01)
        wx.CallAfter(self.onReprocess)

    def onReprocess(self, evt=None, value=None, **kws):
        "handle request reprocess"
        if self.skip_process:
            return
        try:
            dgroup = self.controller.get_group()
        except TypeError:
            return
        if not hasattr(dgroup.config, self.configname):
            return
        form = self.read_form()
        self.process(dgroup=dgroup)
        if self.stale_groups is not None:
            for g in self.stale_groups:
                self.process(dgroup=g, force=True)
            self.stale_groups = None
        self.onPlotEither()


    def process(self, dgroup=None, force_mback=False, force=False, use_form=True, **kws):
        """ handle process (pre-edge/normalize) of XAS data from XAS form
        """
        if self.skip_process and not force:
            return
        if dgroup is None:
            dgroup = self.controller.get_group()
        if dgroup is None:
            return

        self.skip_process = True
        conf = self.get_config(dgroup)
        form = self.read_form()
        if not use_form:
            form.update(self.get_defaultconfig())

        form['group'] = dgroup.groupname

        self.skip_process = False
        scale = form.get('scale', conf.get('scale', 1.0))
        xshift = form.get('xshift', conf.get('xshift', 0.0))
        gname = dgroup.groupname
        cmds = [f"{gname:s}.scale = {scale}",
                f"{gname:s}.xshift = {xshift}",
                f"{gname:s}.xplot = {gname:s}.x+{xshift}",
                f"{gname:s}.ynorm = {gname:s}.y/{scale}",
                f"{gname:s}.dydx  = gradient({gname:s}.ynorm)/gradient({gname:s}.xplot)",
                f"{gname:s}.d2ydx = gradient({gname:s}.dydx)/gradient({gname:s}.xplot)"]

        self.larch_eval('\n'.join(cmds))


        self.unset_skip_process()
        return


    def get_plot_arrays(self, dgroup):
        lab = plotlabels.ynorm
        if dgroup is None:
            return

        dgroup.plot_y2label = None
        dgroup.plot_xlabel = plotlabels.xplot
        dgroup.plot_yarrays = [('y', PLOTOPTS_1, lab)]

        req_attrs = ['y', 'i0', 'ynorm', 'dydx', 'd2ydx']
        pchoice = PlotOne_Choices.get(self.plotone_op.GetStringSelection(), 'ynorm')

        if pchoice in ('y', 'i0', 'ynorm', 'dydx', 'd2ydx'):
            lab = getattr(plotlabels, pchoice)
            dgroup.plot_yarrays = [(pchoice, PLOTOPTS_1, lab)]

        elif pchoice == 'y+dydx':
            lab = plotlabels.y
            dgroup.plot_y2label = lab2 = plotlabels.dydx
            dgroup.plot_yarrays = [('y', PLOTOPTS_1, lab),
                                   ('dydx', PLOTOPTS_D, lab2)]
        elif pchoice == 'ynorm+dydx':
            lab = plotlabels.ynorm
            dgroup.plot_y2label = lab2 = plotlabels.dydx
            dgroup.plot_yarrays = [('ynorm', PLOTOPTS_1, lab),
                                   ('dydx', PLOTOPTS_D, lab2)]

        elif pchoice == 'ynorm+i0':
            lab = plotlabels.ynorm
            dgroup.plot_y2label = lab2 = plotlabels.i0
            dgroup.plot_yarrays = [('ynorm', PLOTOPTS_1, lab),
                                   ('i0', PLOTOPTS_D, lab2)]

        dgroup.plot_ylabel = lab
        needs_proc = False
        for attr in req_attrs:
            needs_proc = needs_proc or (not hasattr(dgroup, attr))

        if needs_proc:
            self.process(dgroup=dgroup, force=True)

        y4e0 = dgroup.yplot = getattr(dgroup, dgroup.plot_yarrays[0][0], dgroup.y)
        dgroup.plot_extras = []

        popts = {'marker': 'o', 'markersize': 5,
                 'label': '_nolegend_',
                 'markerfacecolor': '#888',
                 'markeredgecolor': '#A00'}


    def plot(self, dgroup, title=None, plot_yarrays=None, yoff=0,
             delay_draw=True, multi=False, new=True, with_extras=True, **kws):

        if self.skip_plotting:
            return
        ppanel = self.controller.get_display(stacked=False).panel

        plotcmd = ppanel.oplot
        if new:
            plotcmd = ppanel.plot

        groupname = getattr(dgroup, 'groupname', None)
        if groupname is None:
            return

        if not hasattr(dgroup, 'xplot'):
            print("Cannot plot group ", groupname)

        if ((getattr(dgroup, 'plot_yarrays', None) is None or
             getattr(dgroup, 'dydx', None) is None or
             getattr(dgroup, 'd2ydx', None) is None or
             getattr(dgroup, 'ynorm', None) is None)):
            self.process(dgroup=dgroup)
        self.get_plot_arrays(dgroup)

        if plot_yarrays is None and hasattr(dgroup, 'plot_yarrays'):
            plot_yarrays = dgroup.plot_yarrays

        popts = self.controller.get_plot_conf()
        popts.update(kws)
        popts['grid'] = popts.pop('show_grid')
        popts['fullbox'] = popts.pop('show_fullbox')

        path, fname = path_split(dgroup.filename)
        if 'label' not in popts:
            popts['label'] = dgroup.plot_ylabel

        zoom_limits = get_zoomlimits(ppanel, dgroup)

        popts['xlabel'] = dgroup.plot_xlabel
        popts['ylabel'] = dgroup.plot_ylabel
        if getattr(dgroup, 'plot_y2label', None) is not None:
            popts['y2label'] = dgroup.plot_y2label

        plot_choices = PlotSel_Choices

        if multi:
            ylabel = self.plotsel_op.GetStringSelection()
            yarray_name = plot_choices.get(ylabel, 'ynorm')

            if self.is_xasgroup(dgroup):
                ylabel = getattr(plotlabels, yarray_name, ylabel)
            popts['ylabel'] = ylabel

        plot_extras = None
        if new:
            if title is None:
                title = fname
            plot_extras = getattr(dgroup, 'plot_extras', None)

        popts['title'] = title
        popts['show_legend'] = len(plot_yarrays) > 1
        narr = len(plot_yarrays) - 1

        _linewidth = popts['linewidth']
        for i, pydat in enumerate(plot_yarrays):
            yaname, yopts, yalabel = pydat
            popts.update(yopts)
            if yalabel is not None:
                popts['label'] = yalabel
            linewidht = _linewidth
            if 'linewidth' in popts:
                linewidth = popts.pop('linewidth')
            popts['delay_draw'] = delay_draw

            if yaname == 'i0' and not hasattr(dgroup, yaname):
                dgroup.i0 = np.ones(len(dgroup.xplot))
            plotcmd(dgroup.xplot, getattr(dgroup, yaname)+yoff, linewidth=linewidth, **popts)
            plotcmd = ppanel.oplot

        if with_extras and plot_extras is not None:
            axes = ppanel.axes
            for etype, x, y, opts in plot_extras:
                if etype == 'marker':
                    xpopts = {'marker': 'o', 'markersize': 5,
                              'label': '_nolegend_',
                              'markerfacecolor': 'red',
                              'markeredgecolor': '#884444'}
                    xpopts.update(opts)
                    axes.plot([x], [y], **xpopts)
                elif etype == 'vline':
                    xpopts = {'ymin': 0, 'ymax': 1.0,
                              'label': '_nolegend_',
                              'color': '#888888'}
                    xpopts.update(opts)
                    axes.axvline(x, **xpopts)

        # set_zoomlimits(ppanel, zoom_limits)
        ppanel.reset_formats()
        set_zoomlimits(ppanel, zoom_limits)
        ppanel.conf.unzoom(full=True, delay_draw=False)



from larch import Interpreter
linp = Interpreter()
def onVarChange(group=None, symbolname=None, value=None, **kws):
    print( 'var changed ', group, symbolname, value, kws)

linp('x = 100.0')
linp.symtable.set_symbol('x', 30)
linp.symtable.set_symbol('x', 'a string')
linp('x = arange(7)')



#!/usr/bin/env python
"""
  XAFS pre-edge subtraction, normalization algorithms
"""
import time
from string import printable
from copy import deepcopy
import numpy as np
from lmfit import Parameters, Minimizer, Model
from lmfit.models import (LorentzianModel, GaussianModel, VoigtModel,
                          ConstantModel, LinearModel, QuadraticModel)

from xraydb import guess_edge, xray_edge, core_width

from larch import Group, Make_CallArgs, isgroup, parse_group_args
# now we can reliably import other std and xafs modules...

from larch.math import (index_of, index_nearest, remove_nans2,
                        peak_indices)

from larch.fitting import dict2params
from .xafsutils import set_xafsGroup

@Make_CallArgs(["energy", "norm"])
def prepeaks_setup(energy, norm=None, arrayname=None, group=None, emin=None, emax=None,
                   elo=None, ehi=None, _larch=None):
    """set up pre edge peak group.

    This assumes that pre_edge() has been run successfully on the spectra
    and that the spectra has decent pre-edge subtraction and normalization.

    Arguments:
       energy (ndarray or group): array of x-ray energies, in eV, or group (see note 1)
       norm (ndarray or None):    array of normalized mu(E) to be fit (deprecated, see note 2)
       arrayname (string or None):  name of array to use as data (see note 2)
       group (group or None):     output group
       emax (float or None):      max energy (eV) to use for baesline fit [e0-5]
       emin (float or None):      min energy (eV) to use for baesline fit [e0-40]
       elo: (float or None)       low energy of pre-edge peak region to not fit baseline [e0-20]
       ehi: (float or None)       high energy of pre-edge peak region ot not fit baseline [e0-10]
       _larch (larch instance or None):  current larch session.

    A group named `prepeaks` will be created in the output group, containing:

        ==============   ===========================================================
         attribute        meaning
        ==============   ===========================================================
         energy           energy array for pre-edge peaks = energy[emin:emax]
         norm             spectrum over pre-edge peak energies
        ==============   ===========================================================

    Note that 'norm' will be used here even if a differnt input array was used.

    Notes:
        1. Supports :ref:`First Argument Group` convention, requiring group members `energy` and `norm`
        2. You can pass is an array to fit with 'norm=', or you can name the array to use with
           `arrayname`, which can be one of ['norm', 'flat', 'deconv', 'aspassed', None] with None
           and 'aspassed' meaning that the argument of `norm` will be used, as passed in.
        3. Supports :ref:`Set XAFS Group` convention within Larch or if `_larch` is set.
    """
    ydat = None
    if norm is not None and arrayname in (None, 'aspassed'):
        arrayname = 'aspassed'
        ydat = norm[:]*1.0

    energy, norm, group = parse_group_args(energy, members=('energy', 'norm'),
                                           defaults=(norm,), group=group,
                                           fcn_name='pre_edge_baseline')
    if arrayname == 'flat' and hasattr(group, 'flat'):
        ydat = group.flat[:]*1.0
    elif arrayname == 'deconv' and hasattr(group, 'deconv'):
        ydat = group.deconv[:]*1.0
    if ydat is None:
        ydat = norm[:]*1.0

    if len(energy.shape) > 1:
        energy = energy.squeeze()
    if len(ydat.shape) > 1:
        ydat = ydat.squeeze()

    dat_emin, dat_emax = min(energy), max(energy)
    dat_e0 = getattr(group, 'e0', -1)

    if dat_e0 > 0:
        if emin is None:
            emin = dat_e0 - 30.0
        if emax is None:
            emax = dat_e0 - 1.0
        if elo is None:
            elo = dat_e0 - 15.0
        if ehi is None:
            ehi = dat_e0 - 5.0
        if emin < 0:
            emin += dat_e0
        if elo < 0:
            elo += dat_e0
        if emax < dat_emin:
            emax += dat_e0
        if ehi < dat_emin:
            ehi += dat_e0

    if emax is None or emin is None or elo is None or ehi is None:
        raise ValueError("must provide emin and emax to prepeaks_setup")

    # get indices for input energies
    if emin > emax:
        emin, emax = emax, emin
    if emin > elo:
        elo, emin = emin, elo
    if ehi > emax:
        ehi, emax = emax, ehi

    dele = 1.e-13 + min(np.diff(energy))/5.0

    ilo  = index_of(energy, elo+dele)
    ihi  = index_of(energy, ehi+dele)
    imin = index_of(energy, emin+dele)
    imax = index_of(energy, emax+dele)

    edat = energy[imin: imax+1]
    ydat = ydat[imin:imax+1]

    if not hasattr(group, 'prepeaks'):
        group.prepeaks = Group(energy=edat, norm=ydat,
                               emin=emin, emax=emax,
                               elo=elo, ehi=ehi)
    else:
        group.prepeaks.energy = edat
        group.prepeaks.norm = ydat
        group.prepeaks.emin = emin
        group.prepeaks.emax = emax
        group.prepeaks.elo = elo
        group.prepeaks.ehi = ehi

    group.prepeaks.xplot = edat
    group.prepeaks.yplot = norm
    return

@Make_CallArgs(["energy", "norm"])
def pre_edge_baseline(energy, norm=None, group=None, form='linear+lorentzian',
                      emin=None, emax=None, elo=None, ehi=None, _larch=None):
    """remove baseline from main edge over pre edge peak region

    This assumes that pre_edge() has been run successfully on the spectra
    and that the spectra has decent pre-edge subtraction and normalization.

    Arguments:
       energy (ndarray or group): array of x-ray energies, in eV, or group (see note 1)
       norm (ndarray or group):   array of normalized mu(E)
       group (group or None):     output group
       elo (float or None):       low energy of pre-edge peak region to not fit baseline [e0-20]
       ehi (float or None):       high energy of pre-edge peak region ot not fit baseline [e0-10]
       emax (float or None):      max energy (eV) to use for baesline fit [e0-5]
       emin (float or None):      min energy (eV) to use for baesline fit [e0-40]
       form (string):             form used for baseline (see description)  ['linear+lorentzian']
       _larch (larch instance or None):  current larch session.


    A function will be fit to the input mu(E) data over the range between
    [emin:elo] and [ehi:emax], ignorng the pre-edge peaks in the region
    [elo:ehi].  The baseline function is specified with the `form` keyword
    argument, which can be one or a combination of 'lorentzian', 'gaussian', or 'voigt',
    plus one of 'constant', 'linear', 'quadratic', for example, 'linear+lorentzian',
    'constant+voigt', 'quadratic', 'gaussian'.

    A group named 'prepeaks' will be used or created in the output group, containing

        ==============   ===========================================================
         attribute        meaning
        ==============   ===========================================================
         energy           energy array for pre-edge peaks = energy[emin:emax]
         baseline         fitted baseline array over pre-edge peak energies
         norm             spectrum over pre-edge peak energies
         peaks            baseline-subtraced spectrum over pre-edge peak energies
         centroid         estimated centroid of pre-edge peaks (see note 3)
         peak_energies    list of predicted peak energies (see note 4)
         fit_details      details of fit to extract pre-edge peaks.
        ==============   ===========================================================

    Notes:
       1. Supports :ref:`First Argument Group` convention, requiring group members `energy` and `norm`
       2. Supports :ref:`Set XAFS Group` convention within Larch or if `_larch` is set.
       3. The value calculated for `prepeaks.centroid`  will be found as
          (prepeaks.energy*prepeaks.peaks).sum() / prepeaks.peaks.sum()
       4. The values in the `peak_energies` list will be predicted energies
          of the peaks in `prepeaks.peaks` as found by peakutils.

    """
    energy, norm, group = parse_group_args(energy, members=('energy', 'norm'),
                                           defaults=(norm,), group=group,
                                           fcn_name='pre_edge_baseline')

    prepeaks_setup(energy, norm=norm, group=group, emin=emin, emax=emax,
                   elo=elo, ehi=ehi, _larch=_larch)

    emin = group.prepeaks.emin
    emax = group.prepeaks.emax
    elo = group.prepeaks.elo
    ehi = group.prepeaks.ehi

    dele = 1.e-11 + min(np.diff(energy))/5.0

    imin = index_of(energy, emin+dele)
    ilo  = index_of(energy, elo+dele)
    ihi  = index_of(energy, ehi+dele)
    imax = index_of(energy, emax+dele)

    # build xdat, ydat: dat to fit (skipping pre-edge peaks)
    xdat = np.concatenate((energy[imin:ilo+1], energy[ihi:imax+1]))
    ydat = np.concatenate((norm[imin:ilo+1], norm[ihi:imax+1]))

    edat = energy[imin: imax+1]
    cen = dcen = 0.
    peak_energies = []

    # energy including pre-edge peaks, for output
    norm = norm[imin:imax+1]
    baseline = peaks = dpeaks = norm*0.0

    # build fitting model:
    modelcomps = []
    parvals = {}

    MODELDAT = {'gauss': (GaussianModel, dict(amplitude=1, center=emax, sigma=2)),
                'loren': (LorentzianModel, dict(amplitude=1, center=emax, sigma=2)),
                'voigt': (VoigtModel, dict(amplitude=1, center=emax, sigma=2)),
                'line': (LinearModel, dict(slope=0, intercept=0)),
                'quad': (QuadraticModel, dict(a=0, b=0, c=0)),
                'const': (ConstantModel, dict(c=0))}

    if '+' in form:
        forms = [f.lower() for f in form.split('+')]
    else:
        forms = [form.lower(), '']

    for form in forms[:2]:
        for key, dat in MODELDAT.items():
            if form.startswith(key):
                modelcomps.append(dat[0]())
                parvals.update(dat[1])

    if len(modelcomps) == 0:
        group.prepeaks = Group(energy=edat, norm=norm, baseline=0.0*edat,
                               peaks=0.0*edat, delta_peaks=0.0*edat,
                               centroid=0, delta_centroid=0,
                               peak_energies=[],
                               fit_details=None,
                               emin=emin, emax=emax, elo=elo, ehi=ehi,
                               form=form)
        return

    model = modelcomps.pop()
    if len(modelcomps) > 0:
        model += modelcomps.pop()

    params = model.make_params(**parvals)
    if 'amplitude' in params:
        params['amplitude'].min =  0.0
    if 'sigma' in params:
        params['sigma'].min = 0.05
        params['sigma'].max = 500.0
    if 'center' in params:
        params['center'].max = emax + 25.0
        params['center'].min = emin - 25.0

    result = model.fit(ydat, params, x=xdat)

    # get baseline and resulting norm over edat range
    if result is not None:
        baseline = result.eval(result.params, x=edat)
        peaks = norm-baseline

        # estimate centroid
        cen = (edat*peaks).sum() / peaks.sum()

        # uncertainty in norm includes only uncertainties in baseline fit
        # and uncertainty in centroid:
        try:
            dpeaks = result.eval_uncertainty(result.params, x=edat)
        except:
            dbpeaks = 0.0

        cen_plus = (edat*(peaks+dpeaks)).sum()/ (peaks+dpeaks).sum()
        cen_minus = (edat*(peaks-dpeaks)).sum()/ (peaks-dpeaks).sum()
        dcen = abs(cen_minus - cen_plus) / 2.0

        # locate peak positions
        peak_ids = peak_indices(peaks, threshold=0.05, min_dist=2)
        peak_energies = [edat[pid] for pid in peak_ids]

    group = set_xafsGroup(group, _larch=_larch)
    group.prepeaks = Group(energy=edat, norm=norm, baseline=baseline,
                           peaks=peaks, delta_peaks=dpeaks,
                           centroid=cen, delta_centroid=dcen,
                           peak_energies=peak_energies,
                           fit_details=result,
                           emin=emin, emax=emax, elo=elo, ehi=ehi,
                           form=form)
    return


def prepeaks_fit(group, peakmodel, params, user_options=None, _larch=None):
    """do pre-edge peak fitting - must be done after setting up the fit
    returns a group with Peakfit data, including `result`, the lmfit ModelResult

    """
    prepeaks = getattr(group, 'prepeaks', None)
    if prepeaks is None:
        raise ValueError("must run prepeask_setup() for a group before doing fit")

    if not isinstance(peakmodel, Model):
        raise ValueError("peakmodel must be an lmfit.Model")

    if isinstance(params, dict):
        params = dict2params(params)

    if not isinstance(params, Parameters):
        raise ValueError("params must be an lmfit.Parameters")

    if not hasattr(prepeaks, 'fit_history'):
        prepeaks.fit_history = []

    pkfit = Group()

    for k in ('energy', 'norm', 'norm_std', 'user_options'):
        if hasattr(prepeaks, k):
            setattr(pkfit, k, deepcopy(getattr(prepeaks, k)))

    if user_options is not None:
        pkfit.user_options = user_options

    pkfit.init_fit     = peakmodel.eval(params, x=prepeaks.energy)
    pkfit.init_ycomps  = peakmodel.eval_components(params=params, x=prepeaks.energy)

    norm_std = getattr(prepeaks, 'norm_std', 1.0)
    if isinstance(norm_std, np.ndarray):
        norm_std[np.where(norm_std<1.e-13)] = 1.e-13
    elif norm_std < 0:
        norm_std = 1.0

    pkfit.result = peakmodel.fit(prepeaks.norm, params=params, x=prepeaks.energy,
                                 weights=1.0/norm_std)

    pkfit.ycomps = peakmodel.eval_components(params=pkfit.result.params, x=prepeaks.energy)
    pkfit.label = 'Fit %i' % (1+len(prepeaks.fit_history))

    label = now  = time.strftime("%b-%d %H:%M")
    pkfit.timestamp = time.strftime("%Y-%b-%d %H:%M")
    pkfit.label = label


    fitlabels = [fhist.label for fhist in prepeaks.fit_history]
    if label in fitlabels:
        count = 1
        while label in fitlabels:
            label = f'{now:s}_{printable[count]:s}'
            count +=1
        pkfit.label = label


    prepeaks.fit_history.insert(0, pkfit)
    return pkfit
